@@ -10,7 +10,18 @@ from collections import OrderedDict
 from enum import Enum
 from functools import total_ordering
 from pathlib import Path
-from typing import Any, Callable, Mapping, NamedTuple, Optional, Sequence, Tuple, TypeVar, Union
+from typing import (
+    Any,
+    Callable,
+    List,
+    Mapping,
+    NamedTuple,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+    Union,
+)
 
 _VERSION_PATTERN = r"""
     (?x)                                                (?# ignore whitespace)
@@ -46,7 +57,11 @@ class Vcs(Enum):
 
 
 def _run_cmd(
-    command: str, codes: Sequence[int] = (0,), where: Path = None, shell: bool = False
+    command: str,
+    codes: Sequence[int] = (0,),
+    where: Path = None,
+    shell: bool = False,
+    env: dict = None,
 ) -> Tuple[int, str]:
     result = subprocess.run(
         shlex.split(command),
@@ -54,6 +69,7 @@ def _run_cmd(
         stderr=subprocess.STDOUT,
         cwd=str(where) if where is not None else None,
         shell=shell,
+        env=env,
     )
     output = result.stdout.decode().strip()
     if codes and result.returncode not in codes:
@@ -169,6 +185,97 @@ def _detect_vcs(expected_vcs: Vcs = None) -> Vcs:
         raise RuntimeError("Unable to detect version control system.")
 
 
+class _GitRefInfo:
+    def __init__(
+        self, ref: str, commit: str, creatordate: str, committerdate: str, taggerdate: str
+    ):
+        self.fullref = ref
+        self.commit = commit
+        self.creatordate = self.normalize_git_dt(creatordate)
+        self.committerdate = self.normalize_git_dt(committerdate)
+        self.taggerdate = self.normalize_git_dt(taggerdate)
+        self.tag_topo_lookup = {}  # type: Mapping[str, int]
+
+    def with_tag_topo_lookup(self, lookup: Mapping[str, int]) -> "_GitRefInfo":
+        self.tag_topo_lookup = lookup
+        return self
+
+    @staticmethod
+    def normalize_git_dt(timestamp: str) -> Optional[dt.datetime]:
+        if timestamp == "":
+            return None
+        else:
+            return _parse_git_timestamp_iso_strict(timestamp)
+
+    def __repr__(self):
+        return (
+            "_GitRefInfo(ref={!r}, commit={!r}, creatordate={!r},"
+            " committerdate={!r}, taggerdate={!r})"
+        ).format(
+            self.fullref, self.commit_offset, self.creatordate, self.committerdate, self.taggerdate
+        )
+
+    def best_date(self) -> Optional[dt.datetime]:
+        if self.taggerdate is not None:
+            return self.taggerdate
+        elif self.committerdate is not None:
+            return self.committerdate
+        else:
+            return self.creatordate
+
+    @property
+    def commit_offset(self) -> int:
+        try:
+            return self.tag_topo_lookup[self.fullref]
+        except KeyError:
+            raise RuntimeError(
+                "Unable to determine commit offset for ref {} in data: {}".format(
+                    self.fullref, self.tag_topo_lookup
+                )
+            )
+
+    @property
+    def sort_key(self) -> Tuple[int, Optional[dt.datetime]]:
+        return (-self.commit_offset, self.best_date())
+
+    @property
+    def ref(self) -> str:
+        return self.fullref.replace("refs/tags/", "")
+
+    @staticmethod
+    def normalize_tag_ref(ref: str) -> str:
+        # Older versions of git do not correctly respect --decorate-refs
+        if ref.startswith("refs/tags/"):
+            return ref
+        else:
+            return "refs/tags/{}".format(ref)
+
+    @staticmethod
+    def from_git_tag_topo_order() -> Mapping[str, int]:
+        code, logmsg = _run_cmd(
+            "git log --simplify-by-decoration --topo-order --decorate=full"
+            ' "--decorate-refs=refs/tags/*" HEAD "--format=%H%d"'
+        )
+        tag_lookup = {}
+
+        for tag_offset, line in enumerate(logmsg.strip().splitlines(keepends=False)):
+            # lines have the pattern
+            # <gitsha1>  (tag: refs/tags/v1.2.0b1, tag: refs/tags/v1.2.0)
+            commit, _, tags = line.partition("(")
+            commit = commit.strip()
+            if tags:
+                # remove trailing ')'
+                tags = tags[:-1]
+                taglist = [
+                    tag.strip() for tag in tags.split(",") if tag.strip().startswith("tag: ")
+                ]
+                taglist = [tag.split()[-1] for tag in taglist]
+                taglist = [_GitRefInfo.normalize_tag_ref(tag) for tag in taglist]
+                for tag in taglist:
+                    tag_lookup[tag] = tag_offset
+        return tag_lookup
+
+
 @total_ordering
 class Version:
     def __init__(
@@ -381,7 +488,12 @@ class Version:
 
         code, msg = _run_cmd(
             'git for-each-ref "refs/tags/*" --merged HEAD'
-            ' --format "%(refname)@{%(creatordate:iso-strict)@{%(*committerdate:iso-strict)"'
+            ' --format "%(refname)'
+            "@{%(objectname)"
+            "@{%(creatordate:iso-strict)"
+            "@{%(*committerdate:iso-strict)"
+            "@{%(taggerdate:iso-strict)"
+            '"'
         )
         if not msg:
             try:
@@ -390,20 +502,15 @@ class Version:
             except Exception:
                 distance = 0
             return cls("0.0.0", distance=distance, commit=commit, dirty=dirty)
-        detailed_tags = []
+
+        detailed_tags = []  # type: List[_GitRefInfo]
+        tag_topo_lookup = _GitRefInfo.from_git_tag_topo_order()
+
         for line in msg.strip().splitlines():
             parts = line.split("@{")
-            detailed_tags.append(
-                (
-                    parts[0].replace("refs/tags/", "", 1),
-                    _parse_git_timestamp_iso_strict(parts[1]),
-                    None if parts[2] == "" else _parse_git_timestamp_iso_strict(parts[2]),
-                )
-            )
-        tags = [
-            t[0]
-            for t in reversed(sorted(detailed_tags, key=lambda x: x[1] if x[2] is None else x[2]))
-        ]
+            detailed_tags.append(_GitRefInfo(*parts).with_tag_topo_lookup(tag_topo_lookup))
+
+        tags = [t.ref for t in sorted(detailed_tags, key=lambda x: x.sort_key, reverse=True)]
         tag, base, stage, unmatched, tagged_metadata = _match_version_pattern(
             pattern, tags, latest_tag
         )
@@ -544,10 +651,10 @@ class Version:
         code, msg = _run_cmd("svn status")
         dirty = bool(msg)
 
-        code, msg = _run_cmd("svn info --show-item url")
+        code, msg = _run_cmd("svn info --show-item repos-root-url")
         url = msg.strip("/")
 
-        code, msg = _run_cmd("svn info --show-item last-changed-revision")
+        code, msg = _run_cmd("svn info --show-item revision")
         if not msg or msg == "0":
             commit = None
         else:
